// Code generated by protoc-gen-go. DO NOT EDIT.
// source: drivers/docker/docklog/proto/docker_logger.proto

package proto

import (
	context "context"
	fmt "fmt"
	proto "github.com/golang/protobuf/proto"
	grpc "google.golang.org/grpc"
	codes "google.golang.org/grpc/codes"
	status "google.golang.org/grpc/status"
	math "math"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.ProtoPackageIsVersion3 // please upgrade the proto package

type StartRequest struct {
	Endpoint             string   `protobuf:"bytes,1,opt,name=endpoint,proto3" json:"endpoint,omitempty"`
	ContainerId          string   `protobuf:"bytes,2,opt,name=container_id,json=containerId,proto3" json:"container_id,omitempty"`
	StdoutFifo           string   `protobuf:"bytes,3,opt,name=stdout_fifo,json=stdoutFifo,proto3" json:"stdout_fifo,omitempty"`
	StderrFifo           string   `protobuf:"bytes,4,opt,name=stderr_fifo,json=stderrFifo,proto3" json:"stderr_fifo,omitempty"`
	TlsCert              string   `protobuf:"bytes,5,opt,name=tls_cert,json=tlsCert,proto3" json:"tls_cert,omitempty"`
	TlsKey               string   `protobuf:"bytes,6,opt,name=tls_key,json=tlsKey,proto3" json:"tls_key,omitempty"`
	TlsCa                string   `protobuf:"bytes,7,opt,name=tls_ca,json=tlsCa,proto3" json:"tls_ca,omitempty"`
	Tty                  bool     `protobuf:"varint,8,opt,name=tty,proto3" json:"tty,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *StartRequest) Reset()         { *m = StartRequest{} }
func (m *StartRequest) String() string { return proto.CompactTextString(m) }
func (*StartRequest) ProtoMessage()    {}
func (*StartRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_52d9a4d77fa14396, []int{0}
}

func (m *StartRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StartRequest.Unmarshal(m, b)
}
func (m *StartRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StartRequest.Marshal(b, m, deterministic)
}
func (m *StartRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StartRequest.Merge(m, src)
}
func (m *StartRequest) XXX_Size() int {
	return xxx_messageInfo_StartRequest.Size(m)
}
func (m *StartRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_StartRequest.DiscardUnknown(m)
}

var xxx_messageInfo_StartRequest proto.InternalMessageInfo

func (m *StartRequest) GetEndpoint() string {
	if m != nil {
		return m.Endpoint
	}
	return ""
}

func (m *StartRequest) GetContainerId() string {
	if m != nil {
		return m.ContainerId
	}
	return ""
}

func (m *StartRequest) GetStdoutFifo() string {
	if m != nil {
		return m.StdoutFifo
	}
	return ""
}

func (m *StartRequest) GetStderrFifo() string {
	if m != nil {
		return m.StderrFifo
	}
	return ""
}

func (m *StartRequest) GetTlsCert() string {
	if m != nil {
		return m.TlsCert
	}
	return ""
}

func (m *StartRequest) GetTlsKey() string {
	if m != nil {
		return m.TlsKey
	}
	return ""
}

func (m *StartRequest) GetTlsCa() string {
	if m != nil {
		return m.TlsCa
	}
	return ""
}

func (m *StartRequest) GetTty() bool {
	if m != nil {
		return m.Tty
	}
	return false
}

type StartResponse struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *StartResponse) Reset()         { *m = StartResponse{} }
func (m *StartResponse) String() string { return proto.CompactTextString(m) }
func (*StartResponse) ProtoMessage()    {}
func (*StartResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_52d9a4d77fa14396, []int{1}
}

func (m *StartResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StartResponse.Unmarshal(m, b)
}
func (m *StartResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StartResponse.Marshal(b, m, deterministic)
}
func (m *StartResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StartResponse.Merge(m, src)
}
func (m *StartResponse) XXX_Size() int {
	return xxx_messageInfo_StartResponse.Size(m)
}
func (m *StartResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_StartResponse.DiscardUnknown(m)
}

var xxx_messageInfo_StartResponse proto.InternalMessageInfo

type StopRequest struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *StopRequest) Reset()         { *m = StopRequest{} }
func (m *StopRequest) String() string { return proto.CompactTextString(m) }
func (*StopRequest) ProtoMessage()    {}
func (*StopRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_52d9a4d77fa14396, []int{2}
}

func (m *StopRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StopRequest.Unmarshal(m, b)
}
func (m *StopRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StopRequest.Marshal(b, m, deterministic)
}
func (m *StopRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StopRequest.Merge(m, src)
}
func (m *StopRequest) XXX_Size() int {
	return xxx_messageInfo_StopRequest.Size(m)
}
func (m *StopRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_StopRequest.DiscardUnknown(m)
}

var xxx_messageInfo_StopRequest proto.InternalMessageInfo

type StopResponse struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *StopResponse) Reset()         { *m = StopResponse{} }
func (m *StopResponse) String() string { return proto.CompactTextString(m) }
func (*StopResponse) ProtoMessage()    {}
func (*StopResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_52d9a4d77fa14396, []int{3}
}

func (m *StopResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StopResponse.Unmarshal(m, b)
}
func (m *StopResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StopResponse.Marshal(b, m, deterministic)
}
func (m *StopResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StopResponse.Merge(m, src)
}
func (m *StopResponse) XXX_Size() int {
	return xxx_messageInfo_StopResponse.Size(m)
}
func (m *StopResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_StopResponse.DiscardUnknown(m)
}

var xxx_messageInfo_StopResponse proto.InternalMessageInfo

func init() {
	proto.RegisterType((*StartRequest)(nil), "hashicorp.nomad.drivers.docker.docklog.proto.StartRequest")
	proto.RegisterType((*StartResponse)(nil), "hashicorp.nomad.drivers.docker.docklog.proto.StartResponse")
	proto.RegisterType((*StopRequest)(nil), "hashicorp.nomad.drivers.docker.docklog.proto.StopRequest")
	proto.RegisterType((*StopResponse)(nil), "hashicorp.nomad.drivers.docker.docklog.proto.StopResponse")
}

func init() {
	proto.RegisterFile("drivers/docker/docklog/proto/docker_logger.proto", fileDescriptor_52d9a4d77fa14396)
}

var fileDescriptor_52d9a4d77fa14396 = []byte{
	// 341 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0x9c, 0x50, 0x4f, 0x4f, 0xe2, 0x50,
	0x10, 0xdf, 0x02, 0x2d, 0xdd, 0x01, 0x76, 0x37, 0x2f, 0xd9, 0x58, 0xb9, 0x88, 0x3d, 0x71, 0x30,
	0xc5, 0xe8, 0x49, 0xbc, 0xa9, 0x31, 0x31, 0x7a, 0x82, 0x9b, 0x97, 0xa6, 0xb6, 0x0f, 0x68, 0xa8,
	0x9d, 0x3a, 0x6f, 0x30, 0xe1, 0x64, 0xe2, 0x67, 0xf0, 0x63, 0xfa, 0x21, 0x4c, 0xa7, 0xb5, 0xf1,
	0x0a, 0xa7, 0xf7, 0xe6, 0xf7, 0x2f, 0x33, 0x3f, 0x38, 0x4d, 0x28, 0x7d, 0xd5, 0x64, 0x26, 0x09,
	0xc6, 0x6b, 0x4d, 0xf2, 0x64, 0xb8, 0x9c, 0x14, 0x84, 0x8c, 0x35, 0x18, 0x66, 0xb8, 0x5c, 0x6a,
	0x0a, 0x04, 0x53, 0x27, 0xab, 0xc8, 0xac, 0xd2, 0x18, 0xa9, 0x08, 0x72, 0x7c, 0x8e, 0x92, 0xa0,
	0x4e, 0x08, 0x2a, 0x71, 0x50, 0x27, 0x54, 0x6a, 0xff, 0xd3, 0x82, 0xfe, 0x9c, 0x23, 0xe2, 0x99,
	0x7e, 0xd9, 0x68, 0xc3, 0x6a, 0x08, 0xae, 0xce, 0x93, 0x02, 0xd3, 0x9c, 0x3d, 0x6b, 0x64, 0x8d,
	0x7f, 0xcf, 0x9a, 0x59, 0x1d, 0x43, 0x3f, 0xc6, 0x9c, 0xa3, 0x34, 0xd7, 0x14, 0xa6, 0x89, 0xd7,
	0x12, 0xbe, 0xd7, 0x60, 0x77, 0x89, 0x3a, 0x82, 0x9e, 0xe1, 0x04, 0x37, 0x1c, 0x2e, 0xd2, 0x05,
	0x7a, 0x6d, 0x51, 0x40, 0x05, 0xdd, 0xa6, 0x0b, 0xac, 0x05, 0x9a, 0xa8, 0x12, 0x74, 0x1a, 0x81,
	0x26, 0x12, 0xc1, 0x21, 0xb8, 0x9c, 0x99, 0x30, 0xd6, 0xc4, 0x9e, 0x2d, 0x6c, 0x97, 0x33, 0x73,
	0xad, 0x89, 0xd5, 0x01, 0x94, 0xdf, 0x70, 0xad, 0xb7, 0x9e, 0x23, 0x8c, 0xc3, 0x99, 0xb9, 0xd7,
	0x5b, 0xf5, 0x1f, 0x1c, 0xf1, 0x44, 0x5e, 0x57, 0x70, 0xbb, 0x74, 0x44, 0xea, 0x1f, 0xb4, 0x99,
	0xb7, 0x9e, 0x3b, 0xb2, 0xc6, 0xee, 0xac, 0xfc, 0xfa, 0x7f, 0x61, 0x50, 0x5f, 0x6b, 0x0a, 0xcc,
	0x8d, 0xf6, 0x07, 0xd0, 0x9b, 0x33, 0x16, 0xf5, 0xf5, 0xfe, 0x9f, 0xb2, 0x8d, 0x72, 0xac, 0xe8,
	0xb3, 0x8f, 0x16, 0xf4, 0x6f, 0xa4, 0xb7, 0x07, 0xe9, 0x58, 0xbd, 0x5b, 0x60, 0x4b, 0x82, 0x9a,
	0x06, 0xbb, 0x14, 0x1d, 0xfc, 0x2c, 0x79, 0x78, 0xb9, 0x97, 0xb7, 0x5e, 0xf9, 0x97, 0x7a, 0x83,
	0x4e, 0xb9, 0xa5, 0xba, 0xd8, 0x35, 0xa6, 0x39, 0x74, 0x38, 0xdd, 0xc7, 0xfa, 0xbd, 0xc0, 0x55,
	0xf7, 0xd1, 0x16, 0xfc, 0xc9, 0x91, 0xe7, 0xfc, 0x2b, 0x00, 0x00, 0xff, 0xff, 0xe9, 0x09, 0x25,
	0x99, 0xa7, 0x02, 0x00, 0x00,
}

// Reference imports to suppress errors if they are not otherwise used.
var _ context.Context
var _ grpc.ClientConnInterface

// This is a compile-time assertion to ensure that this generated file
// is compatible with the grpc package it is being compiled against.
const _ = grpc.SupportPackageIsVersion6

// DockerLoggerClient is the client API for DockerLogger service.
//
// For semantics around ctx use and closing/ending streaming RPCs, please refer to https://godoc.org/google.golang.org/grpc#ClientConn.NewStream.
type DockerLoggerClient interface {
	Start(ctx context.Context, in *StartRequest, opts ...grpc.CallOption) (*StartResponse, error)
	Stop(ctx context.Context, in *StopRequest, opts ...grpc.CallOption) (*StopResponse, error)
}

type dockerLoggerClient struct {
	cc grpc.ClientConnInterface
}

func NewDockerLoggerClient(cc grpc.ClientConnInterface) DockerLoggerClient {
	return &dockerLoggerClient{cc}
}

func (c *dockerLoggerClient) Start(ctx context.Context, in *StartRequest, opts ...grpc.CallOption) (*StartResponse, error) {
	out := new(StartResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.drivers.docker.docklog.proto.DockerLogger/Start", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *dockerLoggerClient) Stop(ctx context.Context, in *StopRequest, opts ...grpc.CallOption) (*StopResponse, error) {
	out := new(StopResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.drivers.docker.docklog.proto.DockerLogger/Stop", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

// DockerLoggerServer is the server API for DockerLogger service.
type DockerLoggerServer interface {
	Start(context.Context, *StartRequest) (*StartResponse, error)
	Stop(context.Context, *StopRequest) (*StopResponse, error)
}

// UnimplementedDockerLoggerServer can be embedded to have forward compatible implementations.
type UnimplementedDockerLoggerServer struct {
}

func (*UnimplementedDockerLoggerServer) Start(ctx context.Context, req *StartRequest) (*StartResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Start not implemented")
}
func (*UnimplementedDockerLoggerServer) Stop(ctx context.Context, req *StopRequest) (*StopResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Stop not implemented")
}

func RegisterDockerLoggerServer(s *grpc.Server, srv DockerLoggerServer) {
	s.RegisterService(&_DockerLogger_serviceDesc, srv)
}

func _DockerLogger_Start_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(StartRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(DockerLoggerServer).Start(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.drivers.docker.docklog.proto.DockerLogger/Start",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(DockerLoggerServer).Start(ctx, req.(*StartRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _DockerLogger_Stop_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(StopRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(DockerLoggerServer).Stop(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.drivers.docker.docklog.proto.DockerLogger/Stop",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(DockerLoggerServer).Stop(ctx, req.(*StopRequest))
	}
	return interceptor(ctx, in, info, handler)
}

var _DockerLogger_serviceDesc = grpc.ServiceDesc{
	ServiceName: "hashicorp.nomad.drivers.docker.docklog.proto.DockerLogger",
	HandlerType: (*DockerLoggerServer)(nil),
	Methods: []grpc.MethodDesc{
		{
			MethodName: "Start",
			Handler:    _DockerLogger_Start_Handler,
		},
		{
			MethodName: "Stop",
			Handler:    _DockerLogger_Stop_Handler,
		},
	},
	Streams:  []grpc.StreamDesc{},
	Metadata: "drivers/docker/docklog/proto/docker_logger.proto",
}
