// Code generated by protoc-gen-go. DO NOT EDIT.
// source: drivers/shared/executor/proto/executor.proto

package proto

import (
	context "context"
	fmt "fmt"
	proto "github.com/golang/protobuf/proto"
	timestamp "github.com/golang/protobuf/ptypes/timestamp"
	proto1 "github.com/hashicorp/nomad/plugins/drivers/proto"
	grpc "google.golang.org/grpc"
	codes "google.golang.org/grpc/codes"
	status "google.golang.org/grpc/status"
	math "math"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.ProtoPackageIsVersion3 // please upgrade the proto package

type LaunchRequest struct {
	Cmd                  string                       `protobuf:"bytes,1,opt,name=cmd,proto3" json:"cmd,omitempty"`
	Args                 []string                     `protobuf:"bytes,2,rep,name=args,proto3" json:"args,omitempty"`
	Resources            *proto1.Resources            `protobuf:"bytes,3,opt,name=resources,proto3" json:"resources,omitempty"`
	StdoutPath           string                       `protobuf:"bytes,4,opt,name=stdout_path,json=stdoutPath,proto3" json:"stdout_path,omitempty"`
	StderrPath           string                       `protobuf:"bytes,5,opt,name=stderr_path,json=stderrPath,proto3" json:"stderr_path,omitempty"`
	Env                  []string                     `protobuf:"bytes,6,rep,name=env,proto3" json:"env,omitempty"`
	User                 string                       `protobuf:"bytes,7,opt,name=user,proto3" json:"user,omitempty"`
	TaskDir              string                       `protobuf:"bytes,8,opt,name=task_dir,json=taskDir,proto3" json:"task_dir,omitempty"`
	ResourceLimits       bool                         `protobuf:"varint,9,opt,name=resource_limits,json=resourceLimits,proto3" json:"resource_limits,omitempty"`
	BasicProcessCgroup   bool                         `protobuf:"varint,10,opt,name=basic_process_cgroup,json=basicProcessCgroup,proto3" json:"basic_process_cgroup,omitempty"` // Deprecated: Do not use.
	Mounts               []*proto1.Mount              `protobuf:"bytes,11,rep,name=mounts,proto3" json:"mounts,omitempty"`
	Devices              []*proto1.Device             `protobuf:"bytes,12,rep,name=devices,proto3" json:"devices,omitempty"`
	NetworkIsolation     *proto1.NetworkIsolationSpec `protobuf:"bytes,13,opt,name=network_isolation,json=networkIsolation,proto3" json:"network_isolation,omitempty"`
	NoPivotRoot          bool                         `protobuf:"varint,14,opt,name=no_pivot_root,json=noPivotRoot,proto3" json:"no_pivot_root,omitempty"`
	DefaultPidMode       string                       `protobuf:"bytes,15,opt,name=default_pid_mode,json=defaultPidMode,proto3" json:"default_pid_mode,omitempty"`
	DefaultIpcMode       string                       `protobuf:"bytes,16,opt,name=default_ipc_mode,json=defaultIpcMode,proto3" json:"default_ipc_mode,omitempty"`
	CpusetCgroup         string                       `protobuf:"bytes,17,opt,name=cpuset_cgroup,json=cpusetCgroup,proto3" json:"cpuset_cgroup,omitempty"`
	AllowCaps            []string                     `protobuf:"bytes,18,rep,name=allow_caps,json=allowCaps,proto3" json:"allow_caps,omitempty"`
	Capabilities         []string                     `protobuf:"bytes,19,rep,name=capabilities,proto3" json:"capabilities,omitempty"`
	CgroupV2Override     string                       `protobuf:"bytes,20,opt,name=cgroup_v2_override,json=cgroupV2Override,proto3" json:"cgroup_v2_override,omitempty"`
	CgroupV1Override     map[string]string            `protobuf:"bytes,21,rep,name=cgroup_v1_override,json=cgroupV1Override,proto3" json:"cgroup_v1_override,omitempty" protobuf_key:"bytes,1,opt,name=key,proto3" protobuf_val:"bytes,2,opt,name=value,proto3"`
	OomScoreAdj          int32                        `protobuf:"varint,22,opt,name=oom_score_adj,json=oomScoreAdj,proto3" json:"oom_score_adj,omitempty"`
	WorkDir              string                       `protobuf:"bytes,23,opt,name=work_dir,json=workDir,proto3" json:"work_dir,omitempty"`
	XXX_NoUnkeyedLiteral struct{}                     `json:"-"`
	XXX_unrecognized     []byte                       `json:"-"`
	XXX_sizecache        int32                        `json:"-"`
}

func (m *LaunchRequest) Reset()         { *m = LaunchRequest{} }
func (m *LaunchRequest) String() string { return proto.CompactTextString(m) }
func (*LaunchRequest) ProtoMessage()    {}
func (*LaunchRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{0}
}

func (m *LaunchRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_LaunchRequest.Unmarshal(m, b)
}
func (m *LaunchRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_LaunchRequest.Marshal(b, m, deterministic)
}
func (m *LaunchRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_LaunchRequest.Merge(m, src)
}
func (m *LaunchRequest) XXX_Size() int {
	return xxx_messageInfo_LaunchRequest.Size(m)
}
func (m *LaunchRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_LaunchRequest.DiscardUnknown(m)
}

var xxx_messageInfo_LaunchRequest proto.InternalMessageInfo

func (m *LaunchRequest) GetCmd() string {
	if m != nil {
		return m.Cmd
	}
	return ""
}

func (m *LaunchRequest) GetArgs() []string {
	if m != nil {
		return m.Args
	}
	return nil
}

func (m *LaunchRequest) GetResources() *proto1.Resources {
	if m != nil {
		return m.Resources
	}
	return nil
}

func (m *LaunchRequest) GetStdoutPath() string {
	if m != nil {
		return m.StdoutPath
	}
	return ""
}

func (m *LaunchRequest) GetStderrPath() string {
	if m != nil {
		return m.StderrPath
	}
	return ""
}

func (m *LaunchRequest) GetEnv() []string {
	if m != nil {
		return m.Env
	}
	return nil
}

func (m *LaunchRequest) GetUser() string {
	if m != nil {
		return m.User
	}
	return ""
}

func (m *LaunchRequest) GetTaskDir() string {
	if m != nil {
		return m.TaskDir
	}
	return ""
}

func (m *LaunchRequest) GetResourceLimits() bool {
	if m != nil {
		return m.ResourceLimits
	}
	return false
}

// Deprecated: Do not use.
func (m *LaunchRequest) GetBasicProcessCgroup() bool {
	if m != nil {
		return m.BasicProcessCgroup
	}
	return false
}

func (m *LaunchRequest) GetMounts() []*proto1.Mount {
	if m != nil {
		return m.Mounts
	}
	return nil
}

func (m *LaunchRequest) GetDevices() []*proto1.Device {
	if m != nil {
		return m.Devices
	}
	return nil
}

func (m *LaunchRequest) GetNetworkIsolation() *proto1.NetworkIsolationSpec {
	if m != nil {
		return m.NetworkIsolation
	}
	return nil
}

func (m *LaunchRequest) GetNoPivotRoot() bool {
	if m != nil {
		return m.NoPivotRoot
	}
	return false
}

func (m *LaunchRequest) GetDefaultPidMode() string {
	if m != nil {
		return m.DefaultPidMode
	}
	return ""
}

func (m *LaunchRequest) GetDefaultIpcMode() string {
	if m != nil {
		return m.DefaultIpcMode
	}
	return ""
}

func (m *LaunchRequest) GetCpusetCgroup() string {
	if m != nil {
		return m.CpusetCgroup
	}
	return ""
}

func (m *LaunchRequest) GetAllowCaps() []string {
	if m != nil {
		return m.AllowCaps
	}
	return nil
}

func (m *LaunchRequest) GetCapabilities() []string {
	if m != nil {
		return m.Capabilities
	}
	return nil
}

func (m *LaunchRequest) GetCgroupV2Override() string {
	if m != nil {
		return m.CgroupV2Override
	}
	return ""
}

func (m *LaunchRequest) GetCgroupV1Override() map[string]string {
	if m != nil {
		return m.CgroupV1Override
	}
	return nil
}

func (m *LaunchRequest) GetOomScoreAdj() int32 {
	if m != nil {
		return m.OomScoreAdj
	}
	return 0
}

func (m *LaunchRequest) GetWorkDir() string {
	if m != nil {
		return m.WorkDir
	}
	return ""
}

type LaunchResponse struct {
	Process              *ProcessState `protobuf:"bytes,1,opt,name=process,proto3" json:"process,omitempty"`
	XXX_NoUnkeyedLiteral struct{}      `json:"-"`
	XXX_unrecognized     []byte        `json:"-"`
	XXX_sizecache        int32         `json:"-"`
}

func (m *LaunchResponse) Reset()         { *m = LaunchResponse{} }
func (m *LaunchResponse) String() string { return proto.CompactTextString(m) }
func (*LaunchResponse) ProtoMessage()    {}
func (*LaunchResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{1}
}

func (m *LaunchResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_LaunchResponse.Unmarshal(m, b)
}
func (m *LaunchResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_LaunchResponse.Marshal(b, m, deterministic)
}
func (m *LaunchResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_LaunchResponse.Merge(m, src)
}
func (m *LaunchResponse) XXX_Size() int {
	return xxx_messageInfo_LaunchResponse.Size(m)
}
func (m *LaunchResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_LaunchResponse.DiscardUnknown(m)
}

var xxx_messageInfo_LaunchResponse proto.InternalMessageInfo

func (m *LaunchResponse) GetProcess() *ProcessState {
	if m != nil {
		return m.Process
	}
	return nil
}

type WaitRequest struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *WaitRequest) Reset()         { *m = WaitRequest{} }
func (m *WaitRequest) String() string { return proto.CompactTextString(m) }
func (*WaitRequest) ProtoMessage()    {}
func (*WaitRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{2}
}

func (m *WaitRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_WaitRequest.Unmarshal(m, b)
}
func (m *WaitRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_WaitRequest.Marshal(b, m, deterministic)
}
func (m *WaitRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_WaitRequest.Merge(m, src)
}
func (m *WaitRequest) XXX_Size() int {
	return xxx_messageInfo_WaitRequest.Size(m)
}
func (m *WaitRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_WaitRequest.DiscardUnknown(m)
}

var xxx_messageInfo_WaitRequest proto.InternalMessageInfo

type WaitResponse struct {
	Process              *ProcessState `protobuf:"bytes,1,opt,name=process,proto3" json:"process,omitempty"`
	XXX_NoUnkeyedLiteral struct{}      `json:"-"`
	XXX_unrecognized     []byte        `json:"-"`
	XXX_sizecache        int32         `json:"-"`
}

func (m *WaitResponse) Reset()         { *m = WaitResponse{} }
func (m *WaitResponse) String() string { return proto.CompactTextString(m) }
func (*WaitResponse) ProtoMessage()    {}
func (*WaitResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{3}
}

func (m *WaitResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_WaitResponse.Unmarshal(m, b)
}
func (m *WaitResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_WaitResponse.Marshal(b, m, deterministic)
}
func (m *WaitResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_WaitResponse.Merge(m, src)
}
func (m *WaitResponse) XXX_Size() int {
	return xxx_messageInfo_WaitResponse.Size(m)
}
func (m *WaitResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_WaitResponse.DiscardUnknown(m)
}

var xxx_messageInfo_WaitResponse proto.InternalMessageInfo

func (m *WaitResponse) GetProcess() *ProcessState {
	if m != nil {
		return m.Process
	}
	return nil
}

type ShutdownRequest struct {
	Signal               string   `protobuf:"bytes,1,opt,name=signal,proto3" json:"signal,omitempty"`
	GracePeriod          int64    `protobuf:"varint,2,opt,name=grace_period,json=gracePeriod,proto3" json:"grace_period,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *ShutdownRequest) Reset()         { *m = ShutdownRequest{} }
func (m *ShutdownRequest) String() string { return proto.CompactTextString(m) }
func (*ShutdownRequest) ProtoMessage()    {}
func (*ShutdownRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{4}
}

func (m *ShutdownRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ShutdownRequest.Unmarshal(m, b)
}
func (m *ShutdownRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ShutdownRequest.Marshal(b, m, deterministic)
}
func (m *ShutdownRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ShutdownRequest.Merge(m, src)
}
func (m *ShutdownRequest) XXX_Size() int {
	return xxx_messageInfo_ShutdownRequest.Size(m)
}
func (m *ShutdownRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_ShutdownRequest.DiscardUnknown(m)
}

var xxx_messageInfo_ShutdownRequest proto.InternalMessageInfo

func (m *ShutdownRequest) GetSignal() string {
	if m != nil {
		return m.Signal
	}
	return ""
}

func (m *ShutdownRequest) GetGracePeriod() int64 {
	if m != nil {
		return m.GracePeriod
	}
	return 0
}

type ShutdownResponse struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *ShutdownResponse) Reset()         { *m = ShutdownResponse{} }
func (m *ShutdownResponse) String() string { return proto.CompactTextString(m) }
func (*ShutdownResponse) ProtoMessage()    {}
func (*ShutdownResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{5}
}

func (m *ShutdownResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ShutdownResponse.Unmarshal(m, b)
}
func (m *ShutdownResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ShutdownResponse.Marshal(b, m, deterministic)
}
func (m *ShutdownResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ShutdownResponse.Merge(m, src)
}
func (m *ShutdownResponse) XXX_Size() int {
	return xxx_messageInfo_ShutdownResponse.Size(m)
}
func (m *ShutdownResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_ShutdownResponse.DiscardUnknown(m)
}

var xxx_messageInfo_ShutdownResponse proto.InternalMessageInfo

type UpdateResourcesRequest struct {
	Resources            *proto1.Resources `protobuf:"bytes,1,opt,name=resources,proto3" json:"resources,omitempty"`
	XXX_NoUnkeyedLiteral struct{}          `json:"-"`
	XXX_unrecognized     []byte            `json:"-"`
	XXX_sizecache        int32             `json:"-"`
}

func (m *UpdateResourcesRequest) Reset()         { *m = UpdateResourcesRequest{} }
func (m *UpdateResourcesRequest) String() string { return proto.CompactTextString(m) }
func (*UpdateResourcesRequest) ProtoMessage()    {}
func (*UpdateResourcesRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{6}
}

func (m *UpdateResourcesRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_UpdateResourcesRequest.Unmarshal(m, b)
}
func (m *UpdateResourcesRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_UpdateResourcesRequest.Marshal(b, m, deterministic)
}
func (m *UpdateResourcesRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_UpdateResourcesRequest.Merge(m, src)
}
func (m *UpdateResourcesRequest) XXX_Size() int {
	return xxx_messageInfo_UpdateResourcesRequest.Size(m)
}
func (m *UpdateResourcesRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_UpdateResourcesRequest.DiscardUnknown(m)
}

var xxx_messageInfo_UpdateResourcesRequest proto.InternalMessageInfo

func (m *UpdateResourcesRequest) GetResources() *proto1.Resources {
	if m != nil {
		return m.Resources
	}
	return nil
}

type UpdateResourcesResponse struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *UpdateResourcesResponse) Reset()         { *m = UpdateResourcesResponse{} }
func (m *UpdateResourcesResponse) String() string { return proto.CompactTextString(m) }
func (*UpdateResourcesResponse) ProtoMessage()    {}
func (*UpdateResourcesResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{7}
}

func (m *UpdateResourcesResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_UpdateResourcesResponse.Unmarshal(m, b)
}
func (m *UpdateResourcesResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_UpdateResourcesResponse.Marshal(b, m, deterministic)
}
func (m *UpdateResourcesResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_UpdateResourcesResponse.Merge(m, src)
}
func (m *UpdateResourcesResponse) XXX_Size() int {
	return xxx_messageInfo_UpdateResourcesResponse.Size(m)
}
func (m *UpdateResourcesResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_UpdateResourcesResponse.DiscardUnknown(m)
}

var xxx_messageInfo_UpdateResourcesResponse proto.InternalMessageInfo

type VersionRequest struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *VersionRequest) Reset()         { *m = VersionRequest{} }
func (m *VersionRequest) String() string { return proto.CompactTextString(m) }
func (*VersionRequest) ProtoMessage()    {}
func (*VersionRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{8}
}

func (m *VersionRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_VersionRequest.Unmarshal(m, b)
}
func (m *VersionRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_VersionRequest.Marshal(b, m, deterministic)
}
func (m *VersionRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_VersionRequest.Merge(m, src)
}
func (m *VersionRequest) XXX_Size() int {
	return xxx_messageInfo_VersionRequest.Size(m)
}
func (m *VersionRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_VersionRequest.DiscardUnknown(m)
}

var xxx_messageInfo_VersionRequest proto.InternalMessageInfo

type VersionResponse struct {
	Version              string   `protobuf:"bytes,1,opt,name=version,proto3" json:"version,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *VersionResponse) Reset()         { *m = VersionResponse{} }
func (m *VersionResponse) String() string { return proto.CompactTextString(m) }
func (*VersionResponse) ProtoMessage()    {}
func (*VersionResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{9}
}

func (m *VersionResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_VersionResponse.Unmarshal(m, b)
}
func (m *VersionResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_VersionResponse.Marshal(b, m, deterministic)
}
func (m *VersionResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_VersionResponse.Merge(m, src)
}
func (m *VersionResponse) XXX_Size() int {
	return xxx_messageInfo_VersionResponse.Size(m)
}
func (m *VersionResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_VersionResponse.DiscardUnknown(m)
}

var xxx_messageInfo_VersionResponse proto.InternalMessageInfo

func (m *VersionResponse) GetVersion() string {
	if m != nil {
		return m.Version
	}
	return ""
}

type StatsRequest struct {
	Interval             int64    `protobuf:"varint,1,opt,name=interval,proto3" json:"interval,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *StatsRequest) Reset()         { *m = StatsRequest{} }
func (m *StatsRequest) String() string { return proto.CompactTextString(m) }
func (*StatsRequest) ProtoMessage()    {}
func (*StatsRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{10}
}

func (m *StatsRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StatsRequest.Unmarshal(m, b)
}
func (m *StatsRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StatsRequest.Marshal(b, m, deterministic)
}
func (m *StatsRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StatsRequest.Merge(m, src)
}
func (m *StatsRequest) XXX_Size() int {
	return xxx_messageInfo_StatsRequest.Size(m)
}
func (m *StatsRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_StatsRequest.DiscardUnknown(m)
}

var xxx_messageInfo_StatsRequest proto.InternalMessageInfo

func (m *StatsRequest) GetInterval() int64 {
	if m != nil {
		return m.Interval
	}
	return 0
}

type StatsResponse struct {
	Stats                *proto1.TaskStats `protobuf:"bytes,1,opt,name=stats,proto3" json:"stats,omitempty"`
	XXX_NoUnkeyedLiteral struct{}          `json:"-"`
	XXX_unrecognized     []byte            `json:"-"`
	XXX_sizecache        int32             `json:"-"`
}

func (m *StatsResponse) Reset()         { *m = StatsResponse{} }
func (m *StatsResponse) String() string { return proto.CompactTextString(m) }
func (*StatsResponse) ProtoMessage()    {}
func (*StatsResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{11}
}

func (m *StatsResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_StatsResponse.Unmarshal(m, b)
}
func (m *StatsResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_StatsResponse.Marshal(b, m, deterministic)
}
func (m *StatsResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_StatsResponse.Merge(m, src)
}
func (m *StatsResponse) XXX_Size() int {
	return xxx_messageInfo_StatsResponse.Size(m)
}
func (m *StatsResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_StatsResponse.DiscardUnknown(m)
}

var xxx_messageInfo_StatsResponse proto.InternalMessageInfo

func (m *StatsResponse) GetStats() *proto1.TaskStats {
	if m != nil {
		return m.Stats
	}
	return nil
}

type SignalRequest struct {
	Signal               int32    `protobuf:"varint,1,opt,name=signal,proto3" json:"signal,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *SignalRequest) Reset()         { *m = SignalRequest{} }
func (m *SignalRequest) String() string { return proto.CompactTextString(m) }
func (*SignalRequest) ProtoMessage()    {}
func (*SignalRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{12}
}

func (m *SignalRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_SignalRequest.Unmarshal(m, b)
}
func (m *SignalRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_SignalRequest.Marshal(b, m, deterministic)
}
func (m *SignalRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SignalRequest.Merge(m, src)
}
func (m *SignalRequest) XXX_Size() int {
	return xxx_messageInfo_SignalRequest.Size(m)
}
func (m *SignalRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_SignalRequest.DiscardUnknown(m)
}

var xxx_messageInfo_SignalRequest proto.InternalMessageInfo

func (m *SignalRequest) GetSignal() int32 {
	if m != nil {
		return m.Signal
	}
	return 0
}

type SignalResponse struct {
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *SignalResponse) Reset()         { *m = SignalResponse{} }
func (m *SignalResponse) String() string { return proto.CompactTextString(m) }
func (*SignalResponse) ProtoMessage()    {}
func (*SignalResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{13}
}

func (m *SignalResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_SignalResponse.Unmarshal(m, b)
}
func (m *SignalResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_SignalResponse.Marshal(b, m, deterministic)
}
func (m *SignalResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SignalResponse.Merge(m, src)
}
func (m *SignalResponse) XXX_Size() int {
	return xxx_messageInfo_SignalResponse.Size(m)
}
func (m *SignalResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_SignalResponse.DiscardUnknown(m)
}

var xxx_messageInfo_SignalResponse proto.InternalMessageInfo

type ExecRequest struct {
	Deadline             *timestamp.Timestamp `protobuf:"bytes,1,opt,name=deadline,proto3" json:"deadline,omitempty"`
	Cmd                  string               `protobuf:"bytes,2,opt,name=cmd,proto3" json:"cmd,omitempty"`
	Args                 []string             `protobuf:"bytes,3,rep,name=args,proto3" json:"args,omitempty"`
	XXX_NoUnkeyedLiteral struct{}             `json:"-"`
	XXX_unrecognized     []byte               `json:"-"`
	XXX_sizecache        int32                `json:"-"`
}

func (m *ExecRequest) Reset()         { *m = ExecRequest{} }
func (m *ExecRequest) String() string { return proto.CompactTextString(m) }
func (*ExecRequest) ProtoMessage()    {}
func (*ExecRequest) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{14}
}

func (m *ExecRequest) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ExecRequest.Unmarshal(m, b)
}
func (m *ExecRequest) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ExecRequest.Marshal(b, m, deterministic)
}
func (m *ExecRequest) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ExecRequest.Merge(m, src)
}
func (m *ExecRequest) XXX_Size() int {
	return xxx_messageInfo_ExecRequest.Size(m)
}
func (m *ExecRequest) XXX_DiscardUnknown() {
	xxx_messageInfo_ExecRequest.DiscardUnknown(m)
}

var xxx_messageInfo_ExecRequest proto.InternalMessageInfo

func (m *ExecRequest) GetDeadline() *timestamp.Timestamp {
	if m != nil {
		return m.Deadline
	}
	return nil
}

func (m *ExecRequest) GetCmd() string {
	if m != nil {
		return m.Cmd
	}
	return ""
}

func (m *ExecRequest) GetArgs() []string {
	if m != nil {
		return m.Args
	}
	return nil
}

type ExecResponse struct {
	Output               []byte   `protobuf:"bytes,1,opt,name=output,proto3" json:"output,omitempty"`
	ExitCode             int32    `protobuf:"varint,2,opt,name=exit_code,json=exitCode,proto3" json:"exit_code,omitempty"`
	XXX_NoUnkeyedLiteral struct{} `json:"-"`
	XXX_unrecognized     []byte   `json:"-"`
	XXX_sizecache        int32    `json:"-"`
}

func (m *ExecResponse) Reset()         { *m = ExecResponse{} }
func (m *ExecResponse) String() string { return proto.CompactTextString(m) }
func (*ExecResponse) ProtoMessage()    {}
func (*ExecResponse) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{15}
}

func (m *ExecResponse) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ExecResponse.Unmarshal(m, b)
}
func (m *ExecResponse) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ExecResponse.Marshal(b, m, deterministic)
}
func (m *ExecResponse) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ExecResponse.Merge(m, src)
}
func (m *ExecResponse) XXX_Size() int {
	return xxx_messageInfo_ExecResponse.Size(m)
}
func (m *ExecResponse) XXX_DiscardUnknown() {
	xxx_messageInfo_ExecResponse.DiscardUnknown(m)
}

var xxx_messageInfo_ExecResponse proto.InternalMessageInfo

func (m *ExecResponse) GetOutput() []byte {
	if m != nil {
		return m.Output
	}
	return nil
}

func (m *ExecResponse) GetExitCode() int32 {
	if m != nil {
		return m.ExitCode
	}
	return 0
}

type ProcessState struct {
	Pid                  int32                `protobuf:"varint,1,opt,name=pid,proto3" json:"pid,omitempty"`
	ExitCode             int32                `protobuf:"varint,2,opt,name=exit_code,json=exitCode,proto3" json:"exit_code,omitempty"`
	Signal               int32                `protobuf:"varint,3,opt,name=signal,proto3" json:"signal,omitempty"`
	Time                 *timestamp.Timestamp `protobuf:"bytes,4,opt,name=time,proto3" json:"time,omitempty"`
	OomKilled            bool                 `protobuf:"varint,5,opt,name=oom_killed,json=oomKilled,proto3" json:"oom_killed,omitempty"`
	XXX_NoUnkeyedLiteral struct{}             `json:"-"`
	XXX_unrecognized     []byte               `json:"-"`
	XXX_sizecache        int32                `json:"-"`
}

func (m *ProcessState) Reset()         { *m = ProcessState{} }
func (m *ProcessState) String() string { return proto.CompactTextString(m) }
func (*ProcessState) ProtoMessage()    {}
func (*ProcessState) Descriptor() ([]byte, []int) {
	return fileDescriptor_66b85426380683f3, []int{16}
}

func (m *ProcessState) XXX_Unmarshal(b []byte) error {
	return xxx_messageInfo_ProcessState.Unmarshal(m, b)
}
func (m *ProcessState) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	return xxx_messageInfo_ProcessState.Marshal(b, m, deterministic)
}
func (m *ProcessState) XXX_Merge(src proto.Message) {
	xxx_messageInfo_ProcessState.Merge(m, src)
}
func (m *ProcessState) XXX_Size() int {
	return xxx_messageInfo_ProcessState.Size(m)
}
func (m *ProcessState) XXX_DiscardUnknown() {
	xxx_messageInfo_ProcessState.DiscardUnknown(m)
}

var xxx_messageInfo_ProcessState proto.InternalMessageInfo

func (m *ProcessState) GetPid() int32 {
	if m != nil {
		return m.Pid
	}
	return 0
}

func (m *ProcessState) GetExitCode() int32 {
	if m != nil {
		return m.ExitCode
	}
	return 0
}

func (m *ProcessState) GetSignal() int32 {
	if m != nil {
		return m.Signal
	}
	return 0
}

func (m *ProcessState) GetTime() *timestamp.Timestamp {
	if m != nil {
		return m.Time
	}
	return nil
}

func (m *ProcessState) GetOomKilled() bool {
	if m != nil {
		return m.OomKilled
	}
	return false
}

func init() {
	proto.RegisterType((*LaunchRequest)(nil), "hashicorp.nomad.plugins.executor.proto.LaunchRequest")
	proto.RegisterMapType((map[string]string)(nil), "hashicorp.nomad.plugins.executor.proto.LaunchRequest.CgroupV1OverrideEntry")
	proto.RegisterType((*LaunchResponse)(nil), "hashicorp.nomad.plugins.executor.proto.LaunchResponse")
	proto.RegisterType((*WaitRequest)(nil), "hashicorp.nomad.plugins.executor.proto.WaitRequest")
	proto.RegisterType((*WaitResponse)(nil), "hashicorp.nomad.plugins.executor.proto.WaitResponse")
	proto.RegisterType((*ShutdownRequest)(nil), "hashicorp.nomad.plugins.executor.proto.ShutdownRequest")
	proto.RegisterType((*ShutdownResponse)(nil), "hashicorp.nomad.plugins.executor.proto.ShutdownResponse")
	proto.RegisterType((*UpdateResourcesRequest)(nil), "hashicorp.nomad.plugins.executor.proto.UpdateResourcesRequest")
	proto.RegisterType((*UpdateResourcesResponse)(nil), "hashicorp.nomad.plugins.executor.proto.UpdateResourcesResponse")
	proto.RegisterType((*VersionRequest)(nil), "hashicorp.nomad.plugins.executor.proto.VersionRequest")
	proto.RegisterType((*VersionResponse)(nil), "hashicorp.nomad.plugins.executor.proto.VersionResponse")
	proto.RegisterType((*StatsRequest)(nil), "hashicorp.nomad.plugins.executor.proto.StatsRequest")
	proto.RegisterType((*StatsResponse)(nil), "hashicorp.nomad.plugins.executor.proto.StatsResponse")
	proto.RegisterType((*SignalRequest)(nil), "hashicorp.nomad.plugins.executor.proto.SignalRequest")
	proto.RegisterType((*SignalResponse)(nil), "hashicorp.nomad.plugins.executor.proto.SignalResponse")
	proto.RegisterType((*ExecRequest)(nil), "hashicorp.nomad.plugins.executor.proto.ExecRequest")
	proto.RegisterType((*ExecResponse)(nil), "hashicorp.nomad.plugins.executor.proto.ExecResponse")
	proto.RegisterType((*ProcessState)(nil), "hashicorp.nomad.plugins.executor.proto.ProcessState")
}

func init() {
	proto.RegisterFile("drivers/shared/executor/proto/executor.proto", fileDescriptor_66b85426380683f3)
}

var fileDescriptor_66b85426380683f3 = []byte{
	// 1205 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0xb4, 0x56, 0xfb, 0x6e, 0x1b, 0xc5,
	0x17, 0xee, 0xc6, 0x71, 0x6c, 0x1f, 0xdb, 0x89, 0x3b, 0xbf, 0x36, 0xdd, 0xfa, 0x27, 0xd4, 0xb0,
	0x48, 0xd4, 0x82, 0xb2, 0x69, 0xd3, 0xab, 0x40, 0xa2, 0xd0, 0xa4, 0xa0, 0xaa, 0x17, 0xa2, 0x4d,
	0x69, 0x25, 0xfe, 0x60, 0x99, 0xee, 0x4e, 0xed, 0xa9, 0xd7, 0x3b, 0xcb, 0xcc, 0xac, 0x9b, 0x48,
	0x48, 0xbc, 0x04, 0x48, 0x3c, 0x00, 0xaf, 0xc7, 0x3b, 0xa0, 0xb9, 0x6d, 0xec, 0xb4, 0x54, 0x9b,
	0x22, 0xfe, 0xca, 0x9e, 0xcf, 0xdf, 0xb9, 0xcc, 0x39, 0x73, 0xbe, 0x09, 0x5c, 0x49, 0x39, 0x9d,
	0x13, 0x2e, 0xb6, 0xc5, 0x04, 0x73, 0x92, 0x6e, 0x93, 0x43, 0x92, 0x94, 0x92, 0xf1, 0xed, 0x82,
	0x33, 0xc9, 0x2a, 0x33, 0xd4, 0x26, 0xfa, 0x78, 0x82, 0xc5, 0x84, 0x26, 0x8c, 0x17, 0x61, 0xce,
	0x66, 0x38, 0x0d, 0x8b, 0xac, 0x1c, 0xd3, 0x5c, 0x84, 0xcb, 0xbc, 0xe1, 0xa5, 0x31, 0x63, 0xe3,
	0x8c, 0x98, 0x20, 0x2f, 0xca, 0x97, 0xdb, 0x92, 0xce, 0x88, 0x90, 0x78, 0x56, 0x58, 0x42, 0x60,
	0x1d, 0xb7, 0x5d, 0x7a, 0x93, 0xce, 0x58, 0x86, 0x13, 0xfc, 0xd5, 0x86, 0xfe, 0x23, 0x5c, 0xe6,
	0xc9, 0x24, 0x22, 0x3f, 0x97, 0x44, 0x48, 0x34, 0x80, 0x46, 0x32, 0x4b, 0x7d, 0x6f, 0xcb, 0x1b,
	0x75, 0x22, 0xf5, 0x89, 0x10, 0xac, 0x62, 0x3e, 0x16, 0xfe, 0xca, 0x56, 0x63, 0xd4, 0x89, 0xf4,
	0x37, 0x7a, 0x02, 0x1d, 0x4e, 0x04, 0x2b, 0x79, 0x42, 0x84, 0xdf, 0xd8, 0xf2, 0x46, 0xdd, 0x9d,
	0xab, 0xe1, 0x3f, 0x15, 0x6e, 0xf3, 0x9b, 0x94, 0x61, 0xe4, 0xfc, 0xa2, 0xe3, 0x10, 0xe8, 0x12,
	0x74, 0x85, 0x4c, 0x59, 0x29, 0xe3, 0x02, 0xcb, 0x89, 0xbf, 0xaa, 0xb3, 0x83, 0x81, 0xf6, 0xb1,
	0x9c, 0x58, 0x02, 0xe1, 0xdc, 0x10, 0x9a, 0x15, 0x81, 0x70, 0xae, 0x09, 0x03, 0x68, 0x90, 0x7c,
	0xee, 0xaf, 0xe9, 0x22, 0xd5, 0xa7, 0xaa, 0xbb, 0x14, 0x84, 0xfb, 0x2d, 0xcd, 0xd5, 0xdf, 0xe8,
	0x22, 0xb4, 0x25, 0x16, 0xd3, 0x38, 0xa5, 0xdc, 0x6f, 0x6b, 0xbc, 0xa5, 0xec, 0x3d, 0xca, 0xd1,
	0x65, 0xd8, 0x70, 0xf5, 0xc4, 0x19, 0x9d, 0x51, 0x29, 0xfc, 0xce, 0x96, 0x37, 0x6a, 0x47, 0xeb,
	0x0e, 0x7e, 0xa4, 0x51, 0x74, 0x03, 0xce, 0xbd, 0xc0, 0x82, 0x26, 0x71, 0xc1, 0x59, 0x42, 0x84,
	0x88, 0x93, 0x31, 0x67, 0x65, 0xe1, 0x83, 0x62, 0xdf, 0x5b, 0xf1, 0xbd, 0x08, 0xe9, 0xdf, 0xf7,
	0xcd, 0xcf, 0xbb, 0xfa, 0x57, 0xb4, 0x07, 0x6b, 0x33, 0x56, 0xe6, 0x52, 0xf8, 0xdd, 0xad, 0xc6,
	0xa8, 0xbb, 0x73, 0xa5, 0x66, 0xbb, 0x1e, 0x2b, 0xa7, 0xc8, 0xfa, 0xa2, 0x6f, 0xa1, 0x95, 0x92,
	0x39, 0x55, 0x5d, 0xef, 0xe9, 0x30, 0x9f, 0xd5, 0x0c, 0xb3, 0xa7, 0xbd, 0x22, 0xe7, 0x8d, 0x26,
	0x70, 0x36, 0x27, 0xf2, 0x35, 0xe3, 0xd3, 0x98, 0x0a, 0x96, 0x61, 0x49, 0x59, 0xee, 0xf7, 0xf5,
	0x20, 0xbf, 0xa8, 0x19, 0xf2, 0x89, 0xf1, 0x7f, 0xe0, 0xdc, 0x0f, 0x0a, 0x92, 0x44, 0x83, 0xfc,
	0x04, 0x8a, 0x02, 0xe8, 0xe7, 0x2c, 0x2e, 0xe8, 0x9c, 0xc9, 0x98, 0x33, 0x26, 0xfd, 0x75, 0xdd,
	0xd5, 0x6e, 0xce, 0xf6, 0x15, 0x16, 0x31, 0x26, 0xd1, 0x08, 0x06, 0x29, 0x79, 0x89, 0xcb, 0x4c,
	0xc6, 0x05, 0x4d, 0xe3, 0x19, 0x4b, 0x89, 0xbf, 0xa1, 0xc7, 0xb3, 0x6e, 0xf1, 0x7d, 0x9a, 0x3e,
	0x66, 0x29, 0x59, 0x64, 0xd2, 0x22, 0x31, 0xcc, 0xc1, 0x12, 0xf3, 0x41, 0x91, 0x68, 0xe6, 0x47,
	0xd0, 0x4f, 0x8a, 0x52, 0x10, 0xe9, 0xe6, 0x73, 0x56, 0xd3, 0x7a, 0x06, 0xb4, 0x53, 0xf9, 0x00,
	0x00, 0x67, 0x19, 0x7b, 0x1d, 0x27, 0xb8, 0x10, 0x3e, 0xd2, 0x97, 0xa7, 0xa3, 0x91, 0x5d, 0x5c,
	0x08, 0x14, 0x40, 0x2f, 0xc1, 0x05, 0x7e, 0x41, 0x33, 0x2a, 0x29, 0x11, 0xfe, 0xff, 0x34, 0x61,
	0x09, 0x43, 0x57, 0x00, 0x99, 0x04, 0xf1, 0x7c, 0x27, 0x66, 0x73, 0xc2, 0x39, 0x4d, 0x89, 0x7f,
	0x4e, 0x27, 0x1b, 0x98, 0x5f, 0x9e, 0xed, 0x7c, 0x67, 0x71, 0x74, 0x74, 0xcc, 0xbe, 0x76, 0xcc,
	0x3e, 0xaf, 0x67, 0xf9, 0x30, 0xac, 0xb7, 0xfa, 0xe1, 0xd2, 0xc6, 0x86, 0xe6, 0x28, 0xcf, 0xae,
	0xb9, 0x1c, 0xf7, 0x73, 0xc9, 0x8f, 0xaa, 0xd4, 0x15, 0xac, 0x06, 0xc1, 0xd8, 0x2c, 0x16, 0x09,
	0xe3, 0x24, 0xc6, 0xe9, 0x2b, 0x7f, 0x73, 0xcb, 0x1b, 0x35, 0xa3, 0x2e, 0x63, 0xb3, 0x03, 0x85,
	0x7d, 0x9d, 0xbe, 0x52, 0xfb, 0xa1, 0xef, 0x84, 0xda, 0x8f, 0x0b, 0x66, 0x3f, 0x94, 0xbd, 0x47,
	0xf9, 0x70, 0x17, 0xce, 0xbf, 0x35, 0x93, 0xda, 0xbc, 0x29, 0x39, 0x72, 0x8a, 0x31, 0x25, 0x47,
	0xe8, 0x1c, 0x34, 0xe7, 0x38, 0x2b, 0x89, 0xbf, 0xa2, 0x31, 0x63, 0x7c, 0xbe, 0x72, 0xc7, 0x0b,
	0x7e, 0x82, 0x75, 0x57, 0xbc, 0x28, 0x58, 0x2e, 0x08, 0x7a, 0x02, 0x2d, 0xbb, 0x47, 0x3a, 0x42,
	0x77, 0xe7, 0x46, 0xdd, 0x2e, 0xd8, 0xfd, 0x3a, 0x90, 0x58, 0x92, 0xc8, 0x05, 0x09, 0xfa, 0xd0,
	0x7d, 0x8e, 0xa9, 0xb4, 0xcd, 0x09, 0x7e, 0x84, 0x9e, 0x31, 0xff, 0xa3, 0x74, 0x8f, 0x60, 0xe3,
	0x60, 0x52, 0xca, 0x94, 0xbd, 0xce, 0x9d, 0x82, 0x6e, 0xc2, 0x9a, 0xa0, 0xe3, 0x1c, 0x67, 0xb6,
	0x25, 0xd6, 0x42, 0x1f, 0x42, 0x6f, 0xcc, 0x71, 0x42, 0xe2, 0x82, 0x70, 0xca, 0x52, 0xdd, 0x9c,
	0x46, 0xd4, 0xd5, 0xd8, 0xbe, 0x86, 0x02, 0x04, 0x83, 0xe3, 0x68, 0xa6, 0xe2, 0x60, 0x02, 0x9b,
	0xdf, 0x17, 0xa9, 0x4a, 0x5a, 0x09, 0xa7, 0x4d, 0xb4, 0x24, 0xc2, 0xde, 0xbf, 0x16, 0xe1, 0xe0,
	0x22, 0x5c, 0x78, 0x23, 0x93, 0x2d, 0x62, 0x00, 0xeb, 0xcf, 0x08, 0x17, 0x94, 0xb9, 0x53, 0x06,
	0x9f, 0xc2, 0x46, 0x85, 0xd8, 0xde, 0xfa, 0xd0, 0x9a, 0x1b, 0xc8, 0x9e, 0xdc, 0x99, 0xc1, 0x27,
	0xd0, 0x53, 0x7d, 0xab, 0x2a, 0x1f, 0x42, 0x9b, 0xe6, 0x92, 0xf0, 0xb9, 0x6d, 0x52, 0x23, 0xaa,
	0xec, 0xe0, 0x39, 0xf4, 0x2d, 0xd7, 0x86, 0xfd, 0x06, 0x9a, 0x42, 0x01, 0xa7, 0x3c, 0xe2, 0x53,
	0x2c, 0xa6, 0x26, 0x90, 0x71, 0x0f, 0x2e, 0x43, 0xff, 0x40, 0x4f, 0xe2, 0xed, 0x83, 0x6a, 0xba,
	0x41, 0xa9, 0xc3, 0x3a, 0xa2, 0x3d, 0xfe, 0x14, 0xba, 0xf7, 0x0f, 0x49, 0xe2, 0x1c, 0x6f, 0x41,
	0x3b, 0x25, 0x38, 0xcd, 0x68, 0x4e, 0x6c, 0x51, 0xc3, 0xd0, 0xbc, 0xc6, 0xa1, 0x7b, 0x8d, 0xc3,
	0xa7, 0xee, 0x35, 0x8e, 0x2a, 0xae, 0x7b, 0x5b, 0x57, 0xde, 0x7c, 0x5b, 0x1b, 0xc7, 0x6f, 0x6b,
	0xb0, 0x0b, 0x3d, 0x93, 0xcc, 0x9e, 0x7f, 0x13, 0xd6, 0x58, 0x29, 0x8b, 0x52, 0xea, 0x5c, 0xbd,
	0xc8, 0x5a, 0xe8, 0xff, 0xd0, 0x21, 0x87, 0x54, 0xc6, 0x89, 0xd2, 0xc0, 0x15, 0x7d, 0x82, 0xb6,
	0x02, 0x76, 0x59, 0x4a, 0x82, 0x3f, 0x3d, 0xe8, 0x2d, 0xde, 0x58, 0x95, 0xbb, 0xa0, 0xa9, 0x3d,
	0xa9, 0xfa, 0x7c, 0xa7, 0xff, 0x42, 0x6f, 0x1a, 0x8b, 0xbd, 0x41, 0x21, 0xac, 0xaa, 0xff, 0x33,
	0xf4, 0x0b, 0xfd, 0xee, 0x63, 0x6b, 0x9e, 0x12, 0x58, 0x25, 0x3a, 0x53, 0x9a, 0x65, 0x24, 0xd5,
	0xcf, 0x76, 0x3b, 0xea, 0x30, 0x36, 0x7b, 0xa8, 0x81, 0x9d, 0xdf, 0x3b, 0xd0, 0xbe, 0x6f, 0xf7,
	0x0c, 0x1d, 0xc1, 0x9a, 0x11, 0x07, 0x74, 0xf3, 0xbd, 0x94, 0x70, 0x78, 0xeb, 0xb4, 0x6e, 0x76,
	0xbc, 0x67, 0x90, 0x80, 0x55, 0x25, 0x13, 0xe8, 0x7a, 0xdd, 0x08, 0x0b, 0x1a, 0x33, 0xbc, 0x71,
	0x3a, 0xa7, 0x2a, 0xe9, 0xaf, 0xd0, 0x76, 0xdb, 0x8e, 0x6e, 0xd7, 0x8d, 0x71, 0x42, 0x6d, 0x86,
	0x77, 0x4e, 0xef, 0x58, 0x15, 0xf0, 0x9b, 0x07, 0x1b, 0x27, 0x36, 0x1e, 0x7d, 0x59, 0x37, 0xde,
	0xdb, 0x45, 0x69, 0x78, 0xf7, 0xbd, 0xfd, 0xab, 0xb2, 0x7e, 0x81, 0x96, 0x95, 0x16, 0x54, 0x7b,
	0xa2, 0xcb, 0xea, 0x34, 0xbc, 0x7d, 0x6a, 0xbf, 0x2a, 0xfb, 0x21, 0x34, 0xb5, 0x6c, 0xa0, 0xda,
	0x63, 0x5d, 0x94, 0xb6, 0xe1, 0xcd, 0x53, 0x7a, 0xb9, 0xbc, 0x57, 0x3d, 0x75, 0xff, 0x8d, 0xee,
	0xd4, 0xbf, 0xff, 0x4b, 0x82, 0x56, 0xff, 0xfe, 0x9f, 0x90, 0x37, 0x7d, 0xff, 0xd5, 0x1a, 0xd6,
	0xbf, 0xff, 0x0b, 0x72, 0x58, 0xff, 0xfe, 0x2f, 0xca, 0x5a, 0x70, 0x06, 0xfd, 0xe1, 0x41, 0x5f,
	0x41, 0x07, 0x92, 0x13, 0x3c, 0xa3, 0xf9, 0x18, 0xdd, 0xad, 0xa9, 0xed, 0xca, 0xcb, 0xe8, 0xbb,
	0xf5, 0x74, 0xa5, 0x7c, 0xf5, 0xfe, 0x01, 0x5c, 0x59, 0x23, 0xef, 0xaa, 0x77, 0xaf, 0xf5, 0x43,
	0xd3, 0x48, 0xda, 0x9a, 0xfe, 0x73, 0xfd, 0xef, 0x00, 0x00, 0x00, 0xff, 0xff, 0x5b, 0xc0, 0xa1,
	0x7e, 0xc4, 0x0d, 0x00, 0x00,
}

// Reference imports to suppress errors if they are not otherwise used.
var _ context.Context
var _ grpc.ClientConnInterface

// This is a compile-time assertion to ensure that this generated file
// is compatible with the grpc package it is being compiled against.
const _ = grpc.SupportPackageIsVersion6

// ExecutorClient is the client API for Executor service.
//
// For semantics around ctx use and closing/ending streaming RPCs, please refer to https://godoc.org/google.golang.org/grpc#ClientConn.NewStream.
type ExecutorClient interface {
	Launch(ctx context.Context, in *LaunchRequest, opts ...grpc.CallOption) (*LaunchResponse, error)
	Wait(ctx context.Context, in *WaitRequest, opts ...grpc.CallOption) (*WaitResponse, error)
	Shutdown(ctx context.Context, in *ShutdownRequest, opts ...grpc.CallOption) (*ShutdownResponse, error)
	UpdateResources(ctx context.Context, in *UpdateResourcesRequest, opts ...grpc.CallOption) (*UpdateResourcesResponse, error)
	Version(ctx context.Context, in *VersionRequest, opts ...grpc.CallOption) (*VersionResponse, error)
	Stats(ctx context.Context, in *StatsRequest, opts ...grpc.CallOption) (Executor_StatsClient, error)
	Signal(ctx context.Context, in *SignalRequest, opts ...grpc.CallOption) (*SignalResponse, error)
	Exec(ctx context.Context, in *ExecRequest, opts ...grpc.CallOption) (*ExecResponse, error)
	// buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
	ExecStreaming(ctx context.Context, opts ...grpc.CallOption) (Executor_ExecStreamingClient, error)
}

type executorClient struct {
	cc grpc.ClientConnInterface
}

func NewExecutorClient(cc grpc.ClientConnInterface) ExecutorClient {
	return &executorClient{cc}
}

func (c *executorClient) Launch(ctx context.Context, in *LaunchRequest, opts ...grpc.CallOption) (*LaunchResponse, error) {
	out := new(LaunchResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Launch", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) Wait(ctx context.Context, in *WaitRequest, opts ...grpc.CallOption) (*WaitResponse, error) {
	out := new(WaitResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Wait", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) Shutdown(ctx context.Context, in *ShutdownRequest, opts ...grpc.CallOption) (*ShutdownResponse, error) {
	out := new(ShutdownResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Shutdown", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) UpdateResources(ctx context.Context, in *UpdateResourcesRequest, opts ...grpc.CallOption) (*UpdateResourcesResponse, error) {
	out := new(UpdateResourcesResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/UpdateResources", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) Version(ctx context.Context, in *VersionRequest, opts ...grpc.CallOption) (*VersionResponse, error) {
	out := new(VersionResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Version", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) Stats(ctx context.Context, in *StatsRequest, opts ...grpc.CallOption) (Executor_StatsClient, error) {
	stream, err := c.cc.NewStream(ctx, &_Executor_serviceDesc.Streams[0], "/hashicorp.nomad.plugins.executor.proto.Executor/Stats", opts...)
	if err != nil {
		return nil, err
	}
	x := &executorStatsClient{stream}
	if err := x.ClientStream.SendMsg(in); err != nil {
		return nil, err
	}
	if err := x.ClientStream.CloseSend(); err != nil {
		return nil, err
	}
	return x, nil
}

type Executor_StatsClient interface {
	Recv() (*StatsResponse, error)
	grpc.ClientStream
}

type executorStatsClient struct {
	grpc.ClientStream
}

func (x *executorStatsClient) Recv() (*StatsResponse, error) {
	m := new(StatsResponse)
	if err := x.ClientStream.RecvMsg(m); err != nil {
		return nil, err
	}
	return m, nil
}

func (c *executorClient) Signal(ctx context.Context, in *SignalRequest, opts ...grpc.CallOption) (*SignalResponse, error) {
	out := new(SignalResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Signal", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) Exec(ctx context.Context, in *ExecRequest, opts ...grpc.CallOption) (*ExecResponse, error) {
	out := new(ExecResponse)
	err := c.cc.Invoke(ctx, "/hashicorp.nomad.plugins.executor.proto.Executor/Exec", in, out, opts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *executorClient) ExecStreaming(ctx context.Context, opts ...grpc.CallOption) (Executor_ExecStreamingClient, error) {
	stream, err := c.cc.NewStream(ctx, &_Executor_serviceDesc.Streams[1], "/hashicorp.nomad.plugins.executor.proto.Executor/ExecStreaming", opts...)
	if err != nil {
		return nil, err
	}
	x := &executorExecStreamingClient{stream}
	return x, nil
}

type Executor_ExecStreamingClient interface {
	Send(*proto1.ExecTaskStreamingRequest) error
	Recv() (*proto1.ExecTaskStreamingResponse, error)
	grpc.ClientStream
}

type executorExecStreamingClient struct {
	grpc.ClientStream
}

func (x *executorExecStreamingClient) Send(m *proto1.ExecTaskStreamingRequest) error {
	return x.ClientStream.SendMsg(m)
}

func (x *executorExecStreamingClient) Recv() (*proto1.ExecTaskStreamingResponse, error) {
	m := new(proto1.ExecTaskStreamingResponse)
	if err := x.ClientStream.RecvMsg(m); err != nil {
		return nil, err
	}
	return m, nil
}

// ExecutorServer is the server API for Executor service.
type ExecutorServer interface {
	Launch(context.Context, *LaunchRequest) (*LaunchResponse, error)
	Wait(context.Context, *WaitRequest) (*WaitResponse, error)
	Shutdown(context.Context, *ShutdownRequest) (*ShutdownResponse, error)
	UpdateResources(context.Context, *UpdateResourcesRequest) (*UpdateResourcesResponse, error)
	Version(context.Context, *VersionRequest) (*VersionResponse, error)
	Stats(*StatsRequest, Executor_StatsServer) error
	Signal(context.Context, *SignalRequest) (*SignalResponse, error)
	Exec(context.Context, *ExecRequest) (*ExecResponse, error)
	// buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
	ExecStreaming(Executor_ExecStreamingServer) error
}

// UnimplementedExecutorServer can be embedded to have forward compatible implementations.
type UnimplementedExecutorServer struct {
}

func (*UnimplementedExecutorServer) Launch(ctx context.Context, req *LaunchRequest) (*LaunchResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Launch not implemented")
}
func (*UnimplementedExecutorServer) Wait(ctx context.Context, req *WaitRequest) (*WaitResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Wait not implemented")
}
func (*UnimplementedExecutorServer) Shutdown(ctx context.Context, req *ShutdownRequest) (*ShutdownResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Shutdown not implemented")
}
func (*UnimplementedExecutorServer) UpdateResources(ctx context.Context, req *UpdateResourcesRequest) (*UpdateResourcesResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method UpdateResources not implemented")
}
func (*UnimplementedExecutorServer) Version(ctx context.Context, req *VersionRequest) (*VersionResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Version not implemented")
}
func (*UnimplementedExecutorServer) Stats(req *StatsRequest, srv Executor_StatsServer) error {
	return status.Errorf(codes.Unimplemented, "method Stats not implemented")
}
func (*UnimplementedExecutorServer) Signal(ctx context.Context, req *SignalRequest) (*SignalResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Signal not implemented")
}
func (*UnimplementedExecutorServer) Exec(ctx context.Context, req *ExecRequest) (*ExecResponse, error) {
	return nil, status.Errorf(codes.Unimplemented, "method Exec not implemented")
}
func (*UnimplementedExecutorServer) ExecStreaming(srv Executor_ExecStreamingServer) error {
	return status.Errorf(codes.Unimplemented, "method ExecStreaming not implemented")
}

func RegisterExecutorServer(s *grpc.Server, srv ExecutorServer) {
	s.RegisterService(&_Executor_serviceDesc, srv)
}

func _Executor_Launch_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(LaunchRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Launch(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Launch",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Launch(ctx, req.(*LaunchRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_Wait_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(WaitRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Wait(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Wait",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Wait(ctx, req.(*WaitRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_Shutdown_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ShutdownRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Shutdown(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Shutdown",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Shutdown(ctx, req.(*ShutdownRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_UpdateResources_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(UpdateResourcesRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).UpdateResources(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/UpdateResources",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).UpdateResources(ctx, req.(*UpdateResourcesRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_Version_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(VersionRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Version(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Version",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Version(ctx, req.(*VersionRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_Stats_Handler(srv interface{}, stream grpc.ServerStream) error {
	m := new(StatsRequest)
	if err := stream.RecvMsg(m); err != nil {
		return err
	}
	return srv.(ExecutorServer).Stats(m, &executorStatsServer{stream})
}

type Executor_StatsServer interface {
	Send(*StatsResponse) error
	grpc.ServerStream
}

type executorStatsServer struct {
	grpc.ServerStream
}

func (x *executorStatsServer) Send(m *StatsResponse) error {
	return x.ServerStream.SendMsg(m)
}

func _Executor_Signal_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(SignalRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Signal(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Signal",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Signal(ctx, req.(*SignalRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_Exec_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ExecRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ExecutorServer).Exec(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: "/hashicorp.nomad.plugins.executor.proto.Executor/Exec",
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ExecutorServer).Exec(ctx, req.(*ExecRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _Executor_ExecStreaming_Handler(srv interface{}, stream grpc.ServerStream) error {
	return srv.(ExecutorServer).ExecStreaming(&executorExecStreamingServer{stream})
}

type Executor_ExecStreamingServer interface {
	Send(*proto1.ExecTaskStreamingResponse) error
	Recv() (*proto1.ExecTaskStreamingRequest, error)
	grpc.ServerStream
}

type executorExecStreamingServer struct {
	grpc.ServerStream
}

func (x *executorExecStreamingServer) Send(m *proto1.ExecTaskStreamingResponse) error {
	return x.ServerStream.SendMsg(m)
}

func (x *executorExecStreamingServer) Recv() (*proto1.ExecTaskStreamingRequest, error) {
	m := new(proto1.ExecTaskStreamingRequest)
	if err := x.ServerStream.RecvMsg(m); err != nil {
		return nil, err
	}
	return m, nil
}

var _Executor_serviceDesc = grpc.ServiceDesc{
	ServiceName: "hashicorp.nomad.plugins.executor.proto.Executor",
	HandlerType: (*ExecutorServer)(nil),
	Methods: []grpc.MethodDesc{
		{
			MethodName: "Launch",
			Handler:    _Executor_Launch_Handler,
		},
		{
			MethodName: "Wait",
			Handler:    _Executor_Wait_Handler,
		},
		{
			MethodName: "Shutdown",
			Handler:    _Executor_Shutdown_Handler,
		},
		{
			MethodName: "UpdateResources",
			Handler:    _Executor_UpdateResources_Handler,
		},
		{
			MethodName: "Version",
			Handler:    _Executor_Version_Handler,
		},
		{
			MethodName: "Signal",
			Handler:    _Executor_Signal_Handler,
		},
		{
			MethodName: "Exec",
			Handler:    _Executor_Exec_Handler,
		},
	},
	Streams: []grpc.StreamDesc{
		{
			StreamName:    "Stats",
			Handler:       _Executor_Stats_Handler,
			ServerStreams: true,
		},
		{
			StreamName:    "ExecStreaming",
			Handler:       _Executor_ExecStreaming_Handler,
			ServerStreams: true,
			ClientStreams: true,
		},
	},
	Metadata: "drivers/shared/executor/proto/executor.proto",
}
