//go:build !disable_pgv
// Code generated by protoc-gen-validate. DO NOT EDIT.
// source: envoy/config/core/v3/grpc_method_list.proto

package corev3

import (
	"bytes"
	"errors"
	"fmt"
	"net"
	"net/mail"
	"net/url"
	"regexp"
	"sort"
	"strings"
	"time"
	"unicode/utf8"

	"google.golang.org/protobuf/types/known/anypb"
)

// ensure the imports are used
var (
	_ = bytes.MinRead
	_ = errors.New("")
	_ = fmt.Print
	_ = utf8.UTFMax
	_ = (*regexp.Regexp)(nil)
	_ = (*strings.Reader)(nil)
	_ = net.IPv4len
	_ = time.Duration(0)
	_ = (*url.URL)(nil)
	_ = (*mail.Address)(nil)
	_ = anypb.Any{}
	_ = sort.Sort
)

// Validate checks the field values on GrpcMethodList with the rules defined in
// the proto definition for this message. If any rules are violated, the first
// error encountered is returned, or nil if there are no violations.
func (m *GrpcMethodList) Validate() error {
	return m.validate(false)
}

// ValidateAll checks the field values on GrpcMethodList with the rules defined
// in the proto definition for this message. If any rules are violated, the
// result is a list of violation errors wrapped in GrpcMethodListMultiError,
// or nil if none found.
func (m *GrpcMethodList) ValidateAll() error {
	return m.validate(true)
}

func (m *GrpcMethodList) validate(all bool) error {
	if m == nil {
		return nil
	}

	var errors []error

	for idx, item := range m.GetServices() {
		_, _ = idx, item

		if all {
			switch v := interface{}(item).(type) {
			case interface{ ValidateAll() error }:
				if err := v.ValidateAll(); err != nil {
					errors = append(errors, GrpcMethodListValidationError{
						field:  fmt.Sprintf("Services[%v]", idx),
						reason: "embedded message failed validation",
						cause:  err,
					})
				}
			case interface{ Validate() error }:
				if err := v.Validate(); err != nil {
					errors = append(errors, GrpcMethodListValidationError{
						field:  fmt.Sprintf("Services[%v]", idx),
						reason: "embedded message failed validation",
						cause:  err,
					})
				}
			}
		} else if v, ok := interface{}(item).(interface{ Validate() error }); ok {
			if err := v.Validate(); err != nil {
				return GrpcMethodListValidationError{
					field:  fmt.Sprintf("Services[%v]", idx),
					reason: "embedded message failed validation",
					cause:  err,
				}
			}
		}

	}

	if len(errors) > 0 {
		return GrpcMethodListMultiError(errors)
	}

	return nil
}

// GrpcMethodListMultiError is an error wrapping multiple validation errors
// returned by GrpcMethodList.ValidateAll() if the designated constraints
// aren't met.
type GrpcMethodListMultiError []error

// Error returns a concatenation of all the error messages it wraps.
func (m GrpcMethodListMultiError) Error() string {
	var msgs []string
	for _, err := range m {
		msgs = append(msgs, err.Error())
	}
	return strings.Join(msgs, "; ")
}

// AllErrors returns a list of validation violation errors.
func (m GrpcMethodListMultiError) AllErrors() []error { return m }

// GrpcMethodListValidationError is the validation error returned by
// GrpcMethodList.Validate if the designated constraints aren't met.
type GrpcMethodListValidationError struct {
	field  string
	reason string
	cause  error
	key    bool
}

// Field function returns field value.
func (e GrpcMethodListValidationError) Field() string { return e.field }

// Reason function returns reason value.
func (e GrpcMethodListValidationError) Reason() string { return e.reason }

// Cause function returns cause value.
func (e GrpcMethodListValidationError) Cause() error { return e.cause }

// Key function returns key value.
func (e GrpcMethodListValidationError) Key() bool { return e.key }

// ErrorName returns error name.
func (e GrpcMethodListValidationError) ErrorName() string { return "GrpcMethodListValidationError" }

// Error satisfies the builtin error interface
func (e GrpcMethodListValidationError) Error() string {
	cause := ""
	if e.cause != nil {
		cause = fmt.Sprintf(" | caused by: %v", e.cause)
	}

	key := ""
	if e.key {
		key = "key for "
	}

	return fmt.Sprintf(
		"invalid %sGrpcMethodList.%s: %s%s",
		key,
		e.field,
		e.reason,
		cause)
}

var _ error = GrpcMethodListValidationError{}

var _ interface {
	Field() string
	Reason() string
	Key() bool
	Cause() error
	ErrorName() string
} = GrpcMethodListValidationError{}

// Validate checks the field values on GrpcMethodList_Service with the rules
// defined in the proto definition for this message. If any rules are
// violated, the first error encountered is returned, or nil if there are no violations.
func (m *GrpcMethodList_Service) Validate() error {
	return m.validate(false)
}

// ValidateAll checks the field values on GrpcMethodList_Service with the rules
// defined in the proto definition for this message. If any rules are
// violated, the result is a list of violation errors wrapped in
// GrpcMethodList_ServiceMultiError, or nil if none found.
func (m *GrpcMethodList_Service) ValidateAll() error {
	return m.validate(true)
}

func (m *GrpcMethodList_Service) validate(all bool) error {
	if m == nil {
		return nil
	}

	var errors []error

	if utf8.RuneCountInString(m.GetName()) < 1 {
		err := GrpcMethodList_ServiceValidationError{
			field:  "Name",
			reason: "value length must be at least 1 runes",
		}
		if !all {
			return err
		}
		errors = append(errors, err)
	}

	if len(m.GetMethodNames()) < 1 {
		err := GrpcMethodList_ServiceValidationError{
			field:  "MethodNames",
			reason: "value must contain at least 1 item(s)",
		}
		if !all {
			return err
		}
		errors = append(errors, err)
	}

	if len(errors) > 0 {
		return GrpcMethodList_ServiceMultiError(errors)
	}

	return nil
}

// GrpcMethodList_ServiceMultiError is an error wrapping multiple validation
// errors returned by GrpcMethodList_Service.ValidateAll() if the designated
// constraints aren't met.
type GrpcMethodList_ServiceMultiError []error

// Error returns a concatenation of all the error messages it wraps.
func (m GrpcMethodList_ServiceMultiError) Error() string {
	var msgs []string
	for _, err := range m {
		msgs = append(msgs, err.Error())
	}
	return strings.Join(msgs, "; ")
}

// AllErrors returns a list of validation violation errors.
func (m GrpcMethodList_ServiceMultiError) AllErrors() []error { return m }

// GrpcMethodList_ServiceValidationError is the validation error returned by
// GrpcMethodList_Service.Validate if the designated constraints aren't met.
type GrpcMethodList_ServiceValidationError struct {
	field  string
	reason string
	cause  error
	key    bool
}

// Field function returns field value.
func (e GrpcMethodList_ServiceValidationError) Field() string { return e.field }

// Reason function returns reason value.
func (e GrpcMethodList_ServiceValidationError) Reason() string { return e.reason }

// Cause function returns cause value.
func (e GrpcMethodList_ServiceValidationError) Cause() error { return e.cause }

// Key function returns key value.
func (e GrpcMethodList_ServiceValidationError) Key() bool { return e.key }

// ErrorName returns error name.
func (e GrpcMethodList_ServiceValidationError) ErrorName() string {
	return "GrpcMethodList_ServiceValidationError"
}

// Error satisfies the builtin error interface
func (e GrpcMethodList_ServiceValidationError) Error() string {
	cause := ""
	if e.cause != nil {
		cause = fmt.Sprintf(" | caused by: %v", e.cause)
	}

	key := ""
	if e.key {
		key = "key for "
	}

	return fmt.Sprintf(
		"invalid %sGrpcMethodList_Service.%s: %s%s",
		key,
		e.field,
		e.reason,
		cause)
}

var _ error = GrpcMethodList_ServiceValidationError{}

var _ interface {
	Field() string
	Reason() string
	Key() bool
	Cause() error
	ErrorName() string
} = GrpcMethodList_ServiceValidationError{}
