// Code generated by "stringer -trimprefix=EVENTLOG_LEVEL_ -output strings_event_logger.go -linecomment -type=EventlogLevel"; DO NOT EDIT.

package winsvc

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[EVENTLOG_LEVEL_UNKNOWN-0]
	_ = x[EVENTLOG_LEVEL_INFO-1]
	_ = x[EVENTLOG_LEVEL_WARN-2]
	_ = x[EVENTLOG_LEVEL_ERROR-3]
}

const _EventlogLevel_name = "UNKNOWNINFOWARNERROR"

var _EventlogLevel_index = [...]uint8{0, 7, 11, 15, 20}

func (i EventlogLevel) String() string {
	if i >= EventlogLevel(len(_EventlogLevel_index)-1) {
		return "EventlogLevel(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _EventlogLevel_name[_EventlogLevel_index[i]:_EventlogLevel_index[i+1]]
}
